#include "double.h"
#include "aspen_devel.h"

#if !defined GPU_ARCH
#define	GPU_ARCH	Tesla
#endif

#define	BETA_FIRST	1
#define	UNROLL		8

#define	USE_VOLATILE	0
#define	USE_TEXTURE	0
#define	USE_RESTRICT	0
#define	USE_LDG		0

#define DEFAULT_BLOCK_X		48
#define DEFAULT_BLOCK_Y		5

#if USE_TEXTURE
	DEF_TEXTURE( texture_X, X_ );
	#define	__TRICK_offsetX		,offsetX
#else
	#define	__TRICK_offsetX	
#endif

#define	SUMUP_PREFIX	DGEMV_N
#include "../template/sumup_2D_template.incl"
#include "../template/gemv_n_2D_template.incl"


#define	_CASE_(N,X,Y,A)	\
			case N: API(GEMVN_2D_host) \
					< GPU_ARCH, scalar_t, X, Y, A > \
					( alpha, beta, a, x, y, lda, \
					n1, n2 __TRICK_offsetX ); break;

extern "C" void
API(DGEMV_n_TESLA) ( int blk,
		const scalar_t alpha, const scalar_t beta,
		const scalar_t * a, const scalar_t * x,
		scalar_t * y,
                const int lda,
		const int n1, const int n2 )
{

#if USE_TEXTURE
	size_t offsetX = 0;
	SETUP_TEXTURE ( texture_X, x, n2, offsetX );
#endif

#define	DO(X,Y)	API(GEMVN_2D_host) < GPU_ARCH, scalar_t, X, Y, 8 > \
           	 ( alpha, beta, a, x, y, lda, n1, n2 __TRICK_offsetX )

   if ( n1 <= 1440 ) {
      if ( 1200 <= n1 && n1 <= 1280 ) {
         DO(16,8);
      } else {
         DO(8,10);
      }
   } else if ( n1 <= 1520 ) {
         DO(8,6);
   } else if ( n1 <= 1536 ) {
         DO(16,8);
   } else if ( n1 <= 2872 ) {
      int done = 0;
      if ( 1687 <= n1 && n1 <= 1714 && !done ) { done = 1;
         DO(16,8);
      }
      if ( 2729 <= n1 && n1 <= 2808 && !done ) { done = 1;
         DO(56,3);
      }
      int s4 = (n1-1)%(16*10)+1;
      int s5 = (n1-1)%(24*10)+1;
      int s6 = (n1-1)%(48*10)+1;
      s4 = (16*10)-s4;
      s5 = (24*10)-s5;
      s6 = (48*10)-s6;
      if ( s4 <= s5 && s4 <= s6 && !done ) { done = 1;
         DO(16,4);
      }
      if ( s5 <= s4 && s5 <= s6 && !done ) { done = 1;
         DO(24,6);
      }
      if ( s6 <= s4 && s6 <= s5 && !done ) { done = 1;
         DO(40,4);
      }
   } else {
      if ( 3839 <= n1 && n1 <= 4064 ) {
         DO(16,8);
      } else {
      int s4 = (n1-1)%3840+1;
      if ( s4 <= 160*6 ) {
         int s5 = (n1-1)%2880+1;
         if ( s5 <= 160*6 || lda%256 == 0 ) {
         DO(32,3);
         } else {
         DO(16,5);
         }
      } else {
         DO(16,4);
      }
      }
   }

#if USE_TEXTURE
	cudaUnbindTexture ( texture_X );
#endif

}

