#include "double.h"
#include "aspen_devel.h"


#if !defined(GPU_ARCH)
	#define	GPU_ARCH	(-1)
#endif

#define DEFAULT_BLOCK_X		8


#if GPU_ARCH == Fermi
	#define	GPU_PREFIX	FERMI

	#define	USE_VOLATILE	1
	#define	USE_TEXTURE	1
	#define	USE_RESTRICT	0
	#define	USE_LDG		0

#else
#if GPU_ARCH == Kepler
	#define	GPU_PREFIX	KEPLER

	#define	USE_VOLATILE	1
	#define	USE_TEXTURE	0
	#define	USE_RESTRICT	1
	#define	USE_LDG		0

#else
#if GPU_ARCH == Maxwell
	#define	GPU_PREFIX	MAXWELL

	#define	USE_VOLATILE	1
	#define	USE_TEXTURE	0
	#define	USE_RESTRICT	0
	#define USE_LDG		1

#else
	error !!!
#endif
#endif
#endif


#if USE_TEXTURE
	DEF_TEXTURE ( texture_X, X_ );
	#define	__TRICK_offsetX	,offsetX
#else
	#define	__TRICK_offsetX
#endif


#define	SUMUP_PREFIX		_NAME_1PARAMS_(DGEMVN,GPU_PREFIX)
#define	get_opt_parameter	_NAME_1PARAMS_(SUMUP_PREFIX,Get_Opt_Parameter)
#define	DGEMV_n_STUB		_NAME_1PARAMS_(DGEMV_n,GPU_PREFIX)


#include "../template/sumup_2D_template.incl"
#include "../template/gemv_n-small_template.incl"
#include "../tuning/dgemvn_opt4.h"


#if defined(AT_VERSION)
	#include "../tuning/CURRENT_GPU"
#endif
#if !defined(CURRENT_GPU)
	#define	CURRENT_GPU	(-1)
#endif


#if defined(__CY__)
   #define	_CASE__(x, ... ) \
			_CASE_( x, _BLOCKSIZE_##x, _UX_##x, _MULTI_##x )
#else
   #define	_BLOCKSIZE_(x)		( (((x)/2000)+1)*32 )
   #define	_UX_(x)			( (x%100) )
   #define	_CASE__(x, ... ) \
			_CASE_( x, _BLOCKSIZE_(x), _UX_(x), 1 )
#endif

#define _CASE_(N,X,Y,MULTI)	\
		case N: \
			API(GEMVN_2D_small_host) \
				< GPU_ARCH, scalar_t, X, Y > \
				( alpha, beta, a, x, y, lda, \
					 n1, n2, MULTI __TRICK_offsetX ); break


extern "C" void
API(DGEMV_n_STUB) ( int blk,
		const scalar_t alpha, const scalar_t beta,
		const scalar_t * a, const scalar_t * x,
		scalar_t * y,
                const int lda,
		const int n1, const int n2 )
{

#if USE_TEXTURE
	size_t offsetX = 0;
	SETUP_TEXTURE ( texture_X, x, n2, offsetX );
#endif

#if !defined(AT_VERSION)
	#define	ALL_KERNELS	1
#endif
	int BLK = blk;
	int n = n2;

	#include "../tuning/dgemvn-auto2.h"

	int m = (blk<=0) ? BLK : blk;

#if !defined(AT_VERSION) || CURRENT_GPU==GPU_ARCH
	switch ( m ) {

#if KERNEL_1 || ALL_KERNELS
        _CASE__(     1,  32,  1, 16); // *
#endif
#if KERNEL_101 || ALL_KERNELS
        _CASE__(   101,  32,  1, 15); // *
#endif
#if KERNEL_201 || ALL_KERNELS
        _CASE__(   201,  32,  1, 14); // *
#endif
#if KERNEL_2 || ALL_KERNELS
        _CASE__(     2,  32,  2, 16); // *
#endif
#if KERNEL_102 || ALL_KERNELS
        _CASE__(   102,  32,  2, 15); // *
#endif
#if KERNEL_202 || ALL_KERNELS
        _CASE__(   202,  32,  2, 14); // *
#endif
#if KERNEL_3 || ALL_KERNELS
        _CASE__(     3,  32,  3, 10); // *
#endif
#if KERNEL_103 || ALL_KERNELS
        _CASE__(   103,  32,  3,  9); // *
#endif
#if KERNEL_203 || ALL_KERNELS
        _CASE__(   203,  32,  3,  8); // *
#endif
#if KERNEL_4 || ALL_KERNELS
        _CASE__(     4,  32,  4,  8); // *
#endif
#if KERNEL_104 || ALL_KERNELS
        _CASE__(   104,  32,  4,  7); // *
#endif
#if KERNEL_204 || ALL_KERNELS
        _CASE__(   204,  32,  4,  6); // *
#endif
#if KERNEL_5 || ALL_KERNELS
        _CASE__(     5,  32,  5,  6); // *
#endif
#if KERNEL_105 || ALL_KERNELS
        _CASE__(   105,  32,  5,  5); // *
#endif
#if KERNEL_205 || ALL_KERNELS
        _CASE__(   205,  32,  5,  4); // *
#endif
#if KERNEL_6 || ALL_KERNELS
        _CASE__(     6,  32,  6,  5); // *
#endif
#if KERNEL_106 || ALL_KERNELS
        _CASE__(   106,  32,  6,  4); // *
#endif
#if KERNEL_206 || ALL_KERNELS
        _CASE__(   206,  32,  6,  3); // *
#endif
#if KERNEL_7 || ALL_KERNELS
        _CASE__(     7,  32,  7,  4); // *
#endif
#if KERNEL_107 || ALL_KERNELS
        _CASE__(   107,  32,  7,  3); // *
#endif
#if KERNEL_207 || ALL_KERNELS
        _CASE__(   207,  32,  7,  2); // *
#endif
#if KERNEL_8 || ALL_KERNELS
        _CASE__(     8,  32,  8,  4); // *
#endif
#if KERNEL_108 || ALL_KERNELS
        _CASE__(   108,  32,  8,  3); // *
#endif
#if KERNEL_208 || ALL_KERNELS
        _CASE__(   208,  32,  8,  2); // *
#endif
#if KERNEL_9 || ALL_KERNELS
        _CASE__(     9,  32,  8,  4); // *
#endif
#if KERNEL_109 || ALL_KERNELS
        _CASE__(   109,  32,  8,  3); // *
#endif
#if KERNEL_209 || ALL_KERNELS
        _CASE__(   209,  32,  8,  2); // *
#endif
#if KERNEL_10 || ALL_KERNELS
        _CASE__(    10,  32,  8,  4); // *
#endif
#if KERNEL_110 || ALL_KERNELS
        _CASE__(   110,  32,  8,  3); // *
#endif
#if KERNEL_210 || ALL_KERNELS
        _CASE__(   210,  32,  8,  3); // *
#endif
#if KERNEL_11 || ALL_KERNELS
        _CASE__(    11,  32,  8,  4); // *
#endif
#if KERNEL_111 || ALL_KERNELS
        _CASE__(   111,  32,  8,  3); // *
#endif
#if KERNEL_211 || ALL_KERNELS
        _CASE__(   211,  32,  8,  3); // *
#endif
#if KERNEL_12 || ALL_KERNELS
        _CASE__(    12,  32,  8,  4); // *
#endif
#if KERNEL_112 || ALL_KERNELS
        _CASE__(   112,  32,  8,  3); // *
#endif
#if KERNEL_212 || ALL_KERNELS
        _CASE__(   212,  32,  8,  3); // *
#endif
#if KERNEL_13 || ALL_KERNELS
        _CASE__(    13,  32,  8,  4); // *
#endif
#if KERNEL_113 || ALL_KERNELS
        _CASE__(   113,  32,  8,  3); // *
#endif
#if KERNEL_14 || ALL_KERNELS
        _CASE__(    14,  32,  8,  4); // *
#endif
#if KERNEL_114 || ALL_KERNELS
        _CASE__(   114,  32,  8,  3); // *
#endif
#if KERNEL_15 || ALL_KERNELS
        _CASE__(    15,  32,  8,  4); // *
#endif
#if KERNEL_16 || ALL_KERNELS
        _CASE__(    16,  32,  8,  4); // *
#endif

#if KERNEL_2001 || ALL_KERNELS
        _CASE__(  2001,  64,  1, 16); // *
#endif
#if KERNEL_2002 || ALL_KERNELS
        _CASE__(  2002,  64,  2,  8); // *
#endif
#if KERNEL_2003 || ALL_KERNELS
        _CASE__(  2003,  64,  3,  5); // *
#endif
#if KERNEL_2004 || ALL_KERNELS
        _CASE__(  2004,  64,  4,  4); // *
#endif
#if KERNEL_2005 || ALL_KERNELS
        _CASE__(  2005,  64,  5,  3); // *
#endif
#if KERNEL_2006 || ALL_KERNELS
        _CASE__(  2006,  64,  6,  2); // *
#endif
#if KERNEL_2007 || ALL_KERNELS
        _CASE__(  2007,  64,  7,  2); // *
#endif
#if KERNEL_2008 || ALL_KERNELS
        _CASE__(  2008,  64,  8,  2); // *
#endif
#if KERNEL_2101 || ALL_KERNELS
        _CASE__(  2101,  64,  1, 15); // *
#endif
#if KERNEL_2102 || ALL_KERNELS
        _CASE__(  2102,  64,  2,  7); // *
#endif
#if KERNEL_2103 || ALL_KERNELS
        _CASE__(  2103,  64,  3,  4); // *
#endif

#if KERNEL_4001 || ALL_KERNELS
        _CASE__(  4001,  96,  1, 10); // *
#endif
#if KERNEL_4002 || ALL_KERNELS
        _CASE__(  4002,  96,  2,  5); // x
#endif
#if KERNEL_4003 || ALL_KERNELS
        _CASE__(  4003,  96,  3,  3); // x
#endif
#if KERNEL_4004 || ALL_KERNELS
        _CASE__(  4004,  96,  4,  2); // x
#endif
#if KERNEL_4005 || ALL_KERNELS
        _CASE__(  4005,  96,  5,  2); // x
#endif
#if KERNEL_4006 || ALL_KERNELS
        _CASE__(  4006,  96,  6,  1); // x
#endif

#if KERNEL_6001 || ALL_KERNELS
        _CASE__(  6001, 128,  1,  8); // *
#endif
#if KERNEL_6002 || ALL_KERNELS
        _CASE__(  6002, 128,  2,  4); // xx
#endif
#if KERNEL_6003 || ALL_KERNELS
        _CASE__(  6003, 128,  3,  2); // xx
#endif
#if KERNEL_6004 || ALL_KERNELS
        _CASE__(  6004, 128,  4,  2); // xx
#endif

#if KERNEL_8001 || ALL_KERNELS
        _CASE__(  8001, 160,  1,  6); // xx
#endif
#if KERNEL_8002 || ALL_KERNELS
        _CASE__(  8002, 160,  2,  3); // xx
#endif

#if KERNEL_10001 || ALL_KERNELS
        _CASE__(  10001, 192,  1,  5); // xx
#endif
#if KERNEL_10002 || ALL_KERNELS
        _CASE__(  10002, 192,  2,  2); // xx
#endif

	default:
		fprintf ( stderr,"Not proper NB %d ", m );
		break;
	}
#else
	fprintf( stderr,
		"Caution, ASPEN is tuned up for %s.\n",
		ASPEN_GPU_Names(CURRENT_GPU) );
#endif

#if USE_TEXTURE
	cudaUnbindTexture ( texture_X );
#endif

}

#undef	_CASE_

