#include "double.h"
#include "aspen_devel.h"


#if !defined(GPU_ARCH)
	#define	GPU_ARCH	(-1)
#endif


#if GPU_ARCH == Tesla
	#define	GPU_PREFIX	TESLA

	#define	USE_VOLATILE	1
	#define	USE_TEXTURE	1
	#define	USE_RESTRICT	0
	#define	USE_LDG		0

	#define SPLIT_LOOP	1
#else
#if GPU_ARCH == Fermi
	#define	GPU_PREFIX	FERMI

	#define	USE_VOLATILE	1
	#define	USE_TEXTURE	1
	#define	USE_RESTRICT	0
	#define	USE_LDG		0

	#define SPLIT_LOOP	0
#else
#if GPU_ARCH == Kepler
	#define	GPU_PREFIX	KEPLER

	#define	USE_VOLATILE	1
	#define	USE_TEXTURE	0
	#define	USE_RESTRICT	1
	#define	USE_LDG		0

	#define SPLIT_LOOP	0
#else
#if GPU_ARCH == Maxwell
	#define	GPU_PREFIX	MAXWELL

	#define	USE_VOLATILE	1
	#define	USE_TEXTURE	0
	#define	USE_RESTRICT	0
	#define	USE_LDG		1

	#define SPLIT_LOOP	0
#else
        error !!!
#endif
#endif
#endif
#endif

#if USE_TEXTURE
	DEF_TEXTURE(texture_X, X)
	#define	__TRICK_offsetX	,offsetX
#else
	#define	__TRICK_offsetX
#endif

#define	SUMUP_PREFIX		_NAME_1PARAMS_(DGEMVT,GPU_PREFIX)
#define	do_best_kernel_DGEMVT	_NAME_1PARAMS_(SUMUP_PREFIX,Do_Best_Kernel)
#define	DGEMV_t_STUB		_NAME_1PARAMS_(DGEMV_t,GPU_PREFIX)

#include "../template/sumup_1Dx2_template.incl"
#include "../template/gemv_t_template.incl"


#if defined(AT_VERSION)
	#include "../tuning/CURRENT_GPU"
#endif
#if !defined(CURRENT_GPU)
	#define	CURRENT_GPU	(-1)
#endif

#if defined(AT_VERSION)
#if CURRENT_GPU==GPU_ARCH
	#include "../tuning/dgemvt-at-opt.c"
#endif
#else
#if 1
	#include "../tuning/dgemvt-at-opt.c"
#else
long
do_best_kernel_DGEMVT ( int blk,
		const scalar_t alpha, const scalar_t beta,
		const scalar_t * a, const scalar_t * x,
		scalar_t * y,
		int lda,
		int n1, int n2, int offsetX,
		double * ff )
{
	GEMVT_KERNEL < GPU_ARCH, scalar_t, 32, 1, 1, 1 >
		( alpha, beta, a, x, y, lda, n1, n2, offsetX );
	return (long)0;
}
#endif
#endif


extern "C" __host__ void
API(DGEMV_t_STUB)( int blk,
		const scalar_t alpha, const scalar_t beta,
		const scalar_t * a, const scalar_t * x,
		scalar_t * y,
                const int lda,
		const int n1, const int n2 )
{
// y = alpha * A^t * x + beta * y
// A(1:n2,1:n1) :: A^T(1:n1,1:n2)
// y(1:n1)
// x(1:n2)
   if ( n1 <= 0 || n2 <= 0 ) return;
   if ( alpha == (scalar_t)0 && beta == (scalar_t)1 ) return;

   size_t offsetX = 0;

#if USE_TEXTURE
   SETUP_TEXTURE ( texture_X, x, n2, offsetX );
#endif

#if !defined(AT_VERSION) || CURRENT_GPU==GPU_ARCH
   double f;
   do_best_kernel_DGEMVT ( blk, alpha, beta, a, x, y, lda, n1, n2,
				offsetX, &f );
#else
	fprintf( stderr,
		"Caution, ASPEN is tuned up for %s.\n",
		ASPEN_GPU_Names(CURRENT_GPU) );
#endif

#if USE_TEXTURE
   cudaUnbindTexture ( texture_X );
#endif

}

