#include "double.h"
#include "aspen_devel.h"


extern "C" void
API(DGEMV_t) ( int blk,
		const scalar_t alpha, const scalar_t beta,
		const scalar_t * a, const scalar_t * x,
		scalar_t * y,
                const int lda,
		const int n1, const int n2 )
{
// y = alpha * A^t * x + beta * y
// A(1:n2,1:n1) :: A^T(1:n1,1:n2)
// y(1:n1)
// x(1:n2)
   if ( n1 <= 0 || n2 <= 0 ) return;
   if ( alpha == (scalar_t)0 && beta == (scalar_t)1 ) return;

   int Ver = API(get_device_Major) ( );

   switch ( Ver ) {
   case Tesla:
	API(DGEMV_t_TESLA) ( blk,
		alpha, beta, a, x, y, lda, n1, n2 );
	break;
   case Fermi:
	API(DGEMV_t_FERMI) ( blk,
		alpha, beta, a, x, y, lda, n1, n2 );
	break;
   case Kepler:
	API(DGEMV_t_KEPLER) ( blk,
		alpha, beta, a, x, y, lda, n1, n2 );
	break;
   case Maxwell:
	API(DGEMV_t_MAXWELL) ( blk,
		alpha, beta, a, x, y, lda, n1, n2 );
	break;
   }

}

