#include "float.h"
#include "aspen_devel.h"


#define	GPU_ARCH	Tesla

#define	BETA_FIRST	1

#define	USE_VOLATILE	0
#define	USE_TEXTURE	0
#define	USE_RESTRICT	0
#define	USE_LDG		0


#if USE_TEXTURE
	DEF_TEXTURE(texture_X, X_)
	#define	__TRICK_offsetX		,offsetX
#else
	#define	__TRICK_offsetX	
#endif

#define DEFAULT_BLOCK_X		64
#define DEFAULT_BLOCK_Y		4


#define SUMUP_PREFIX    SGEMV_NN
#include "../template/sumup_2D_template.incl"
#define UNROLL  4
#define	atomicDoubleAdd	atomicDoubleAdd_
#define	USE_ATOMIC_ADD	0
#include "../template/gemv_n_2D_template.incl"



#define _DO_(X,Y)	\
			API(GEMVN_2D_host) < GPU_ARCH, scalar_t, X, Y, 8 > \
					( alpha, beta, a, x, y, lda, \
					 n1, n2 __TRICK_offsetX )
#define _CASE_(N,X,Y)	\
			case N: _DO_(X,Y); break


extern "C" void
API(SGEMV_n_TESLA) ( int blk,
		const scalar_t alpha, const scalar_t beta,
		const scalar_t * a, const scalar_t * x,
		scalar_t * y,
                const int lda,
		const int n1, const int n2 )
{

#if USE_TEXTURE
	size_t offsetX = 0;
	SETUP_TEXTURE(texture_X, x, n2, offsetX);
#endif

	int m;

        if ( blk <= 0 ) {
		if ( n1 <= 906	) { _DO_( 32,16); }
		else if ( n1 <= 1408 ) { _DO_( 48,10); }
		else if ( n1 <= 1920 ) { _DO_( 96, 5); }
		else if ( n1 <= 2560 ) { _DO_(128, 4); }
		else if ( n1 <= 2880 ) { _DO_( 96, 5); }
		else if ( n1 <= 3840 ) { _DO_(128, 4); }
		else if ( n1 <= 4032 ) { _DO_(224, 2); }
		else if ( n1 <= 4096 ) { _DO_(256, 2); }
		else if ( n1 <= 4480 ) { _DO_(224, 2); }
		else if ( n1 <= 5248 ) { _DO_(256, 2); }
		else if ( n1 <= 5696 ) {
			if ( lda % 128 == 0 ) { _DO_(128, 1); }
			else { _DO_(192, 2); }
		}
		else if ( n1 <= 6208 ) { _DO_(128, 1); }
		else if ( n1 <= 6592 ) { _DO_(224, 2); }
		else if ( n1 <= 7808 ) { _DO_(256, 1); }
		else if ( n1 <= 9024 ) { _DO_(128, 1); }
		else if ( n1 <=10368 ) { _DO_(256, 1); }
		else if ( n1 <=11456 ) { _DO_(128, 1); }
		else if ( n1 <=12928 ) { _DO_(256, 1); }
		else { _DO_(128, 1); }
	} else {	

		m = blk;
		switch (m) {
		_CASE_(  1, 32, 1);
		_CASE_(  2, 64, 1);
		_CASE_(  3, 96, 1);
		_CASE_(  4,128, 1);
		_CASE_(  5,160, 1);
		_CASE_(  6,192, 1);
		_CASE_(  7,224, 1);
		_CASE_(  8,256, 1);

		_CASE_( 32, 32,16);
		_CASE_( 48, 48,10);
		_CASE_( 64, 64, 8);
		_CASE_( 96, 96, 5);
		_CASE_(128,128, 4);
		_CASE_(160,160, 3);
		_CASE_(192,192, 2);
		_CASE_(224,224, 2);
		_CASE_(256,256, 2);

		default:
		fprintf ( stderr,"Not proper NB %d ", m );
		break;
		}

	}

#if USE_TEXTURE
	cudaUnbindTexture ( texture_X );
#endif

}

