#include "float.h"
#include "aspen_devel.h"

#if !defined(GPU_ARCH)
	#define	GPU_ARCH	(-1)
#endif

#define	NUM_THREADS		192
#define	SPLIT_LOOP		0	/* meaningless in this kernel */
#define	LOOP_RANGE_ARRAY	1

#if GPU_ARCH == Fermi
	#define GPU_PREFIX	FERMI

	#define	USE_VOLATILE	1
	#define	USE_TEXTURE	1
	#define	USE_RESTRICT	0
	#define	USE_LDG		0
#else
#if GPU_ARCH == Kepler
	#define GPU_PREFIX	KEPLER

	#define	USE_VOLATILE	1
	#define	USE_TEXTURE	0
	#define	USE_RESTRICT	1
	#define	USE_LDG		0
#else
#if GPU_ARCH == Maxwell
	#define GPU_PREFIX	MAXWELL

	#define	USE_VOLATILE	1
	#define	USE_TEXTURE	0
	#define	USE_RESTRICT	0
	#define	USE_LDG		1
#else
	error !!
#endif
#endif
#endif


#if USE_TEXTURE
	DEF_TEXTURE( texture_X, X );
#endif

#define	SUMUP_PREFIX	_NAME_1PARAMS_(SSYMVU,GPU_PREFIX)
#define	SSYMVu_STUB	_NAME_1PARAMS_(SSYMVu_LARGE,GPU_PREFIX)

#include "../template/sumup_1Dx2_template.incl"
#include "../template/symv_upper_atomic_template.incl"

#if defined(AT_VERSION)
	#include "../tuning/CURRENT_GPU"
#endif
#if !defined(CURRENT_GPU)
	#define	CURRENT_GPU	(-1)
#endif


extern "C" void
API(SSYMVu_STUB) ( const int blk,
         const int n, const scalar_t * a, const int lda,
         const scalar_t * x, scalar_t * y,
         const scalar_t alpha, const scalar_t beta )
{

// y = alpha * A * x + beta * y
// y(1:n1)
// x(1:n2)

#if defined(AT_VERSION)
#if CURRENT_GPU==GPU_ARCH
      int BLK = blk;
	#include "../tuning/param-ssymv.h"
	#include "../tuning/ssymv-upper-auto.h"
#else
	fprintf( stderr,
		"Caution, ASPEN is tuned up for %s.\n",
		ASPEN_GPU_Names(CURRENT_GPU) );
#endif
#else
	API(SYMVu_ATOMIC_host)
		< GPU_ARCH, scalar_t, 128, 1, 16, 7, 0 >
		( n, a, lda, x, y, alpha, beta );
#endif

}

