#if GPU_ARCH>=Kepler

template < class TYPE >
__forceinline__ __device__ TYPE
__SHFL_xor ( TYPE const x, int const laneMask, int const width )
{
      return  __shfl_xor( x, laneMask, width );
}

template < >
__forceinline__ __device__ double
__SHFL_xor < double > ( double const x, int const laneMask, int const width )
{
      int tmp_hi, tmp_lo;
#if 1
      asm volatile ( "mov.b64  {%1, %0}, %2;\n\t"
			:"=r"(tmp_hi), "=r"(tmp_lo)
			:"d"(x) );
#else
      tmp_hi = __double2hiint( x );
      tmp_lo = __double2loint( x );
#endif
      tmp_hi = __shfl_xor( tmp_hi, laneMask, width );
      tmp_lo = __shfl_xor( tmp_lo, laneMask, width );
      return  __hiloint2double( tmp_hi, tmp_lo );
}

template < class TYPE >
__forceinline__ __device__ TYPE
half_warp_Reduce ( TYPE x )
{

      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      x += __SHFL_xor < TYPE > ( x,  2, 32 );
      x += __SHFL_xor < TYPE > ( x,  4, 32 );
      x += __SHFL_xor < TYPE > ( x,  8, 32 );

      return x;
}

template < class TYPE >
__forceinline__ __device__ TYPE
half_warp_Reduce_2 ( TYPE x, TYPE y )
{

      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      y += __SHFL_xor < TYPE > ( y,  1, 32 );
      x = ((threadIdx.x & 1) == 0) ? x : y; //xyxyxyxyxyxyxyxy
      x += __SHFL_xor < TYPE > ( x,  2, 32 );
      x += __SHFL_xor < TYPE > ( x,  4, 32 );
      x += __SHFL_xor < TYPE > ( x,  8, 32 );

      return x;
}

template < class TYPE >
__forceinline__ __device__ TYPE
half_warp_Reduce_3 ( TYPE x, TYPE y, TYPE z )
{

      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      y += __SHFL_xor < TYPE > ( y,  1, 32 );
      x = ((threadIdx.x & 1) == 0) ? x : y; //xyxyxyxyxyxyxyxy
      z += __SHFL_xor < TYPE > ( z,  1, 32 );
      x += __SHFL_xor < TYPE > ( x,  2, 32 );
      z += __SHFL_xor < TYPE > ( z,  2, 32 );
      x = ((threadIdx.x & 2) == 0) ? x : z; //xyzzxyzzxyzzxyzz
      x += __SHFL_xor < TYPE > ( x,  4, 32 );
      x += __SHFL_xor < TYPE > ( x,  8, 32 );

      return x;
}

template < class TYPE >
__forceinline__ __device__ TYPE
half_warp_Reduce_4 ( TYPE x, TYPE y, TYPE z, TYPE w )
{

      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      y += __SHFL_xor < TYPE > ( y,  1, 32 );
      x = ((threadIdx.x & 1) == 0) ? x : y; //xyxyxyxyxyxyxyxy
      x += __SHFL_xor < TYPE > ( x,  2, 32 );
      z += __SHFL_xor < TYPE > ( z,  1, 32 );
      w += __SHFL_xor < TYPE > ( w,  1, 32 );
      z = ((threadIdx.x & 1) == 0) ? z : w; //zwzwzwzwzwzwzwzw
      z += __SHFL_xor < TYPE > ( z,  2, 32 );
      x = ((threadIdx.x & 2) == 0) ? x : z; //xyzwxyzwxyzwxyzw
      x += __SHFL_xor < TYPE > ( x,  4, 32 );
      x += __SHFL_xor < TYPE > ( x,  8, 32 );

      return x;
}

#if 0
template < class TYPE >
__forceinline__ __device__ TYPE
half_warp_Reduce_8 ( TYPE x, TYPE y, TYPE z, TYPE w, TYPE a, TYPE b, TYPE c, TYPE d )
{

      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      y += __SHFL_xor < TYPE > ( y,  1, 32 );
      z += __SHFL_xor < TYPE > ( z,  1, 32 );
      w += __SHFL_xor < TYPE > ( w,  1, 32 );
      a += __SHFL_xor < TYPE > ( a,  1, 32 );
      b += __SHFL_xor < TYPE > ( b,  1, 32 );
      c += __SHFL_xor < TYPE > ( c,  1, 32 );
      d += __SHFL_xor < TYPE > ( d,  1, 32 );
      x = ((threadIdx.x & 1) == 0) ? x : y; //xyxyxyxyxyxyxyxy
      z = ((threadIdx.x & 1) == 0) ? z : w; //zwzwzwzwzwzwzwzw
      a = ((threadIdx.x & 1) == 0) ? a : b; //abababababababab
      c = ((threadIdx.x & 1) == 0) ? c : d; //cdcdcdcdcdcdcdcd
      x += __SHFL_xor < TYPE > ( x,  2, 32 );
      z += __SHFL_xor < TYPE > ( z,  2, 32 );
      a += __SHFL_xor < TYPE > ( a,  2, 32 );
      c += __SHFL_xor < TYPE > ( c,  2, 32 );
      x = ((threadIdx.x & 2) == 0) ? x : z; //xyzwxyzwxyzwxyzw
      a = ((threadIdx.x & 2) == 0) ? a : c; //abcdabcdabcdabcd
      x += __SHFL_xor < TYPE > ( x,  4, 32 );
      a += __SHFL_xor < TYPE > ( a,  4, 32 );
      x = ((threadIdx.x & 4) == 0) ? x : a; //xyzwabcdxyzwabcd
      x += __SHFL_xor < TYPE > ( x,  8, 32 );

      return x;
}
#endif

template < class TYPE, int LANES >
__forceinline__ __device__ TYPE
warp_Reduce ( TYPE x )
{

if ( LANES > 16 ) { // 5
      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      x += __SHFL_xor < TYPE > ( x,  2, 32 );
      x += __SHFL_xor < TYPE > ( x,  4, 32 );
      x += __SHFL_xor < TYPE > ( x,  8, 32 );
      x += __SHFL_xor < TYPE > ( x, 16, 32 );
} else
if ( LANES >  8 ) { // 4
      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      x += __SHFL_xor < TYPE > ( x,  2, 32 );
      x += __SHFL_xor < TYPE > ( x,  4, 32 );
      x += __SHFL_xor < TYPE > ( x,  8, 32 );
} else
if ( LANES >  4 ) { // 3
      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      x += __SHFL_xor < TYPE > ( x,  2, 32 );
      x += __SHFL_xor < TYPE > ( x,  4, 32 );
} else
if ( LANES >  2 ) { // 2
      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      x += __SHFL_xor < TYPE > ( x,  2, 32 );
} else
                  { // 1
      x += __SHFL_xor < TYPE > ( x,  1, 32 );
}

      return x;
}

template < class TYPE, int LANES >
__forceinline__ __device__ TYPE
warp_Reduce_2 ( TYPE x, TYPE y )
{

if ( LANES > 16 ) { // 7
      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      y += __SHFL_xor < TYPE > ( y,  1, 32 );
      x = ((threadIdx.x & 1) == 0) ? x : y; //xyxyxyxyxyxyxyxyxyxyxyxyxyxyxyxy
      x += __SHFL_xor < TYPE > ( x,  2, 32 );
      x += __SHFL_xor < TYPE > ( x,  4, 32 );
      x += __SHFL_xor < TYPE > ( x,  8, 32 );
      x += __SHFL_xor < TYPE > ( x, 16, 32 );
} else
if ( LANES >  8 ) { // 6
      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      y += __SHFL_xor < TYPE > ( y,  1, 32 );
      x = ((threadIdx.x & 1) == 0) ? x : y; //xyxyxyxyxyxyxyxy0000000000000000
      x += __SHFL_xor < TYPE > ( x,  2, 32 );
      x += __SHFL_xor < TYPE > ( x,  4, 32 );
      x += __SHFL_xor < TYPE > ( x,  8, 32 );
} else
if ( LANES >  4 ) { // 5
      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      y += __SHFL_xor < TYPE > ( y,  1, 32 );
      x = ((threadIdx.x & 1) == 0) ? x : y; //xyxyxyxy000000000000000000000000
      x += __SHFL_xor < TYPE > ( x,  2, 32 );
      x += __SHFL_xor < TYPE > ( x,  4, 32 );
} else
if ( LANES >  2 ) { // 4
      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      y += __SHFL_xor < TYPE > ( y,  1, 32 );
      x = ((threadIdx.x & 1) == 0) ? x : y; //xyxy0000000000000000000000000000
      x += __SHFL_xor < TYPE > ( x,  2, 32 );
} else
                  { // 3
      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      y += __SHFL_xor < TYPE > ( y,  1, 32 );
      x = ((threadIdx.x & 1) == 0) ? x : y; //xy000000000000000000000000000000
}

      return x;
}

template < class TYPE, int LANES >
__forceinline__ __device__ TYPE
warp_Reduce_3 ( TYPE x, TYPE y, TYPE z )
{

if ( LANES > 16 ) { // 12
      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      y += __SHFL_xor < TYPE > ( y,  1, 32 );
      x = ((threadIdx.x & 1) == 0) ? x : y; //xyxyxyxyxyxyxyxyxyxyxyxyxyxyxyxy
      z += __SHFL_xor < TYPE > ( z,  1, 32 );
      x += __SHFL_xor < TYPE > ( x,  2, 32 );
      z += __SHFL_xor < TYPE > ( z,  2, 32 );
      x = ((threadIdx.x & 2) == 0) ? x : z; //xyzzxyzzxyzzxyzzxyzzxyzzxyzzxyzz
      x += __SHFL_xor < TYPE > ( x,  4, 32 );
      x += __SHFL_xor < TYPE > ( x,  8, 32 );
      x += __SHFL_xor < TYPE > ( x, 16, 32 );
} else
if ( LANES >  8 ) { // 11
      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      y += __SHFL_xor < TYPE > ( y,  1, 32 );
      x = ((threadIdx.x & 1) == 0) ? x : y; //xyxyxyxyxyxyxyxy0000000000000000
      z += __SHFL_xor < TYPE > ( z,  1, 32 );
      x += __SHFL_xor < TYPE > ( x,  2, 32 );
      z += __SHFL_xor < TYPE > ( z,  2, 32 );
      x = ((threadIdx.x & 2) == 0) ? x : z; //xyzzxyzzxyzzxyzz0000000000000000
      x += __SHFL_xor < TYPE > ( x,  4, 32 );
      x += __SHFL_xor < TYPE > ( x,  8, 32 );
} else
if ( LANES >  4 ) { // 10
      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      y += __SHFL_xor < TYPE > ( y,  1, 32 );
      x = ((threadIdx.x & 1) == 0) ? x : y; //xyxyxyxy000000000000000000000000
      z += __SHFL_xor < TYPE > ( z,  1, 32 );
      x += __SHFL_xor < TYPE > ( x,  2, 32 );
      z += __SHFL_xor < TYPE > ( z,  2, 32 );
      x = ((threadIdx.x & 2) == 0) ? x : z; //xyzzxyzz000000000000000000000000
      x += __SHFL_xor < TYPE > ( x,  4, 32 );
} else
if ( LANES >  2 ) { // 9
      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      y += __SHFL_xor < TYPE > ( y,  1, 32 );
      x = ((threadIdx.x & 1) == 0) ? x : y; //xyxy0000000000000000000000000000
      z += __SHFL_xor < TYPE > ( z,  1, 32 );
      x += __SHFL_xor < TYPE > ( x,  2, 32 );
      z += __SHFL_xor < TYPE > ( z,  2, 32 );
      x = ((threadIdx.x & 2) == 0) ? x : z; //xyzz0000000000000000000000000000
} else
                   { // 7
      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      y += __SHFL_xor < TYPE > ( y,  1, 32 );
      x = ((threadIdx.x & 1) == 0) ? x : y; //xy000000000000000000000000000000
      z += __SHFL_xor < TYPE > ( z,  1, 32 );
      x += __SHFL_xor < TYPE > ( z,  2, 32);//xyzz0000000000000000000000000000
}

      return x;
}

template < class TYPE, int LANES >
__forceinline__ __device__ TYPE
warp_Reduce_4 ( TYPE x, TYPE y, TYPE z, TYPE w )
{

if ( LANES > 16 ) { // 12
      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      y += __SHFL_xor < TYPE > ( y,  1, 32 );
      x = ((threadIdx.x & 1) == 0) ? x : y; //xyxyxyxyxyxyxyxyxyxyxyxyxyxyxyxy
      x += __SHFL_xor < TYPE > ( x,  2, 32 );
      z += __SHFL_xor < TYPE > ( z,  1, 32 );
      w += __SHFL_xor < TYPE > ( w,  1, 32 );
      z = ((threadIdx.x & 1) == 0) ? z : w; //zwzwzwzwzwzwzwzwzwzwzwzwzwzwzwzw
      z += __SHFL_xor < TYPE > ( z,  2, 32 );
      x = ((threadIdx.x & 2) == 0) ? x : z; //xyzwxyzwxyzwxyzwxyzwxyzwxyzwxyzw
      x += __SHFL_xor < TYPE > ( x,  4, 32 );
      x += __SHFL_xor < TYPE > ( x,  8, 32 );
      x += __SHFL_xor < TYPE > ( x, 16, 32 );
} else
if ( LANES >  8 ) { // 11
      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      y += __SHFL_xor < TYPE > ( y,  1, 32 );
      x = ((threadIdx.x & 1) == 0) ? x : y; //xyxyxyxyxyxyxyxy0000000000000000
      x += __SHFL_xor < TYPE > ( x,  2, 32 );
      z += __SHFL_xor < TYPE > ( z,  1, 32 );
      w += __SHFL_xor < TYPE > ( w,  1, 32 );
      z = ((threadIdx.x & 1) == 0) ? z : w; //zwzwzwzwzwzwzwzw0000000000000000
      z += __SHFL_xor < TYPE > ( z,  2, 32 );
      x = ((threadIdx.x & 2) == 0) ? x : z; //xyzwxyzwxyzwxyzw0000000000000000
      x += __SHFL_xor < TYPE > ( x,  4, 32 );
      x += __SHFL_xor < TYPE > ( x,  8, 32 );
} else
if ( LANES >  4 ) { // 10
      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      y += __SHFL_xor < TYPE > ( y,  1, 32 );
      x = ((threadIdx.x & 1) == 0) ? x : y; //xyxyxyxy000000000000000000000000
      x += __SHFL_xor < TYPE > ( x,  2, 32 );
      z += __SHFL_xor < TYPE > ( z,  1, 32 );
      w += __SHFL_xor < TYPE > ( w,  1, 32 );
      z = ((threadIdx.x & 1) == 0) ? z : w; //zwzwzwzw000000000000000000000000
      z += __SHFL_xor < TYPE > ( z,  2, 32 );
      x = ((threadIdx.x & 2) == 0) ? x : z; //xyzwxyzw000000000000000000000000
      x += __SHFL_xor < TYPE > ( x,  4, 32 );
} else
if ( LANES >  2 ) { // 9
      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      y += __SHFL_xor < TYPE > ( y,  1, 32 );
      x = ((threadIdx.x & 1) == 0) ? x : y; //xyxy0000000000000000000000000000
      x += __SHFL_xor < TYPE > ( x,  2, 32 );
      z += __SHFL_xor < TYPE > ( z,  1, 32 );
      w += __SHFL_xor < TYPE > ( w,  1, 32 );
      z = ((threadIdx.x & 1) == 0) ? z : w; //zwzw0000000000000000000000000000
      z += __SHFL_xor < TYPE > ( z,  2, 32 );
      x = ((threadIdx.x & 2) == 0) ? x : z; //xyzw0000000000000000000000000000
} else
                   { // 7
      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      y += __SHFL_xor < TYPE > ( y,  1, 32 );
      x = ((threadIdx.x & 1) == 0) ? x : y; //xy000000000000000000000000000000
      z += __SHFL_xor < TYPE > ( z,  1, 32 );
      w += __SHFL_xor < TYPE > ( w,  1, 32 );
      z = ((threadIdx.x & 1) == 0) ? z : w; //zw000000000000000000000000000000
      x += __SHFL_xor < TYPE > ( z,  2, 32);//xyzw0000000000000000000000000000
}

      return x;
}

#if 0
template < class TYPE, int LANES >
__forceinline__ __device__ TYPE
warp_Reduce_8 ( TYPE x, TYPE y, TYPE z, TYPE w, TYPE a, TYPE b, TYPE c, TYPE d )
{

if ( LANES > 16 ) { // 12
      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      y += __SHFL_xor < TYPE > ( y,  1, 32 );
      x = ((threadIdx.x & 1) == 0) ? x : y; //xyxyxyxyxyxyxyxyxyxyxyxyxyxyxyxy
      z += __SHFL_xor < TYPE > ( z,  1, 32 );
      w += __SHFL_xor < TYPE > ( w,  1, 32 );
      z = ((threadIdx.x & 1) == 0) ? z : w; //zwzwzwzwzwzwzwzwzwzwzwzwzwzwzwzw
      a += __SHFL_xor < TYPE > ( a,  1, 32 );
      b += __SHFL_xor < TYPE > ( b,  1, 32 );
      a = ((threadIdx.x & 1) == 0) ? a : b; //abababababababababababababababab
      c += __SHFL_xor < TYPE > ( c,  1, 32 );
      d += __SHFL_xor < TYPE > ( d,  1, 32 );
      c = ((threadIdx.x & 1) == 0) ? c : d; //cdcdcdcdcdcdcdcdcdcdcdcdcdcdcdcd
      x += __SHFL_xor < TYPE > ( x,  2, 32 );
      z += __SHFL_xor < TYPE > ( z,  2, 32 );
      x = ((threadIdx.x & 2) == 0) ? x : z; //xyzwxyzwxyzwxyzwxyzwxyzwxyzwxyzw
      a += __SHFL_xor < TYPE > ( a,  2, 32 );
      c += __SHFL_xor < TYPE > ( c,  2, 32 );
      a = ((threadIdx.x & 2) == 0) ? a : c; //abcdabcdabcdabcdabcdabcdabcdabcd
      x += __SHFL_xor < TYPE > ( x,  4, 32 );
      a += __SHFL_xor < TYPE > ( a,  4, 32 );
      x = ((threadIdx.x & 4) == 0) ? x : a; //xyzwabcdxyzwabcdxyzwabcdxyzwabcd
      x += __SHFL_xor < TYPE > ( x,  8, 32 );
      x += __SHFL_xor < TYPE > ( x, 16, 32 );
} else
if ( LANES >  8 ) { // 11
      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      y += __SHFL_xor < TYPE > ( y,  1, 32 );
      x = ((threadIdx.x & 1) == 0) ? x : y; //xyxyxyxyxyxyxyxy0000000000000000
      z += __SHFL_xor < TYPE > ( z,  1, 32 );
      w += __SHFL_xor < TYPE > ( w,  1, 32 );
      z = ((threadIdx.x & 1) == 0) ? z : w; //zwzwzwzwzwzwzwzw0000000000000000
      a += __SHFL_xor < TYPE > ( a,  1, 32 );
      b += __SHFL_xor < TYPE > ( b,  1, 32 );
      a = ((threadIdx.x & 1) == 0) ? a : b; //abababababababab0000000000000000
      c += __SHFL_xor < TYPE > ( c,  1, 32 );
      d += __SHFL_xor < TYPE > ( d,  1, 32 );
      c = ((threadIdx.x & 1) == 0) ? c : d; //cdcdcdcdcdcdcdcd0000000000000000
      x += __SHFL_xor < TYPE > ( x,  2, 32 );
      z += __SHFL_xor < TYPE > ( z,  2, 32 );
      x = ((threadIdx.x & 2) == 0) ? x : z; //xyzwxyzwxyzwxyzw0000000000000000
      a += __SHFL_xor < TYPE > ( a,  2, 32 );
      c += __SHFL_xor < TYPE > ( c,  2, 32 );
      a = ((threadIdx.x & 2) == 0) ? a : c; //abcdabcdabcdabcd0000000000000000
      x += __SHFL_xor < TYPE > ( x,  4, 32 );
      a += __SHFL_xor < TYPE > ( a,  4, 32 );
      x = ((threadIdx.x & 4) == 0) ? x : a; //xyzwabcdxyzwabcd0000000000000000
      x += __SHFL_xor < TYPE > ( x,  8, 32 );
} else
if ( LANES >  4 ) { // 10
      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      y += __SHFL_xor < TYPE > ( y,  1, 32 );
      x = ((threadIdx.x & 1) == 0) ? x : y; //xyxyxyxy000000000000000000000000
      z += __SHFL_xor < TYPE > ( z,  1, 32 );
      w += __SHFL_xor < TYPE > ( w,  1, 32 );
      z = ((threadIdx.x & 1) == 0) ? z : w; //zwzwzwzw000000000000000000000000
      a += __SHFL_xor < TYPE > ( a,  1, 32 );
      b += __SHFL_xor < TYPE > ( b,  1, 32 );
      a = ((threadIdx.x & 1) == 0) ? a : b; //abababab000000000000000000000000
      c += __SHFL_xor < TYPE > ( c,  1, 32 );
      d += __SHFL_xor < TYPE > ( d,  1, 32 );
      c = ((threadIdx.x & 1) == 0) ? c : d; //cdcdcdcd000000000000000000000000
      x += __SHFL_xor < TYPE > ( x,  2, 32 );
      z += __SHFL_xor < TYPE > ( z,  2, 32 );
      x = ((threadIdx.x & 2) == 0) ? x : z; //xyzwxyzw000000000000000000000000
      a += __SHFL_xor < TYPE > ( a,  2, 32 );
      c += __SHFL_xor < TYPE > ( c,  2, 32 );
      a = ((threadIdx.x & 2) == 0) ? a : c; //abcdabcd000000000000000000000000
      x += __SHFL_xor < TYPE > ( x,  4, 32 );
      a += __SHFL_xor < TYPE > ( a,  4, 32 );
      x = ((threadIdx.x & 4) == 0) ? x : a; //xyzwabcd000000000000000000000000
} else
if ( LANES >  2 ) { // 9
      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      y += __SHFL_xor < TYPE > ( y,  1, 32 );
      x = ((threadIdx.x & 1) == 0) ? x : y; //xyxy0000000000000000000000000000
      z += __SHFL_xor < TYPE > ( z,  1, 32 );
      w += __SHFL_xor < TYPE > ( w,  1, 32 );
      z = ((threadIdx.x & 1) == 0) ? z : w; //zwzw0000000000000000000000000000
      a += __SHFL_xor < TYPE > ( a,  1, 32 );
      b += __SHFL_xor < TYPE > ( b,  1, 32 );
      a = ((threadIdx.x & 1) == 0) ? a : b; //abab0000000000000000000000000000
      c += __SHFL_xor < TYPE > ( c,  1, 32 );
      d += __SHFL_xor < TYPE > ( d,  1, 32 );
      c = ((threadIdx.x & 1) == 0) ? c : d; //cdcd0000000000000000000000000000
      x += __SHFL_xor < TYPE > ( x,  2, 32 );
      z += __SHFL_xor < TYPE > ( z,  2, 32 );
      x = ((threadIdx.x & 2) == 0) ? x : z; //xyzw0000000000000000000000000000
      a += __SHFL_xor < TYPE > ( a,  2, 32 );
      c += __SHFL_xor < TYPE > ( c,  2, 32 );
      a = ((threadIdx.x & 2) == 0) ? a : c; //abcd0000000000000000000000000000
      x += __SHFL_xor < TYPE > ( a,  4, 32);//xyzwabcd000000000000000000000000
} else
                   { // 7
      x += __SHFL_xor < TYPE > ( x,  1, 32 );
      y += __SHFL_xor < TYPE > ( y,  1, 32 );
      x = ((threadIdx.x & 1) == 0) ? x : y; //xy000000000000000000000000000000
      z += __SHFL_xor < TYPE > ( z,  1, 32 );
      w += __SHFL_xor < TYPE > ( w,  1, 32 );
      z = ((threadIdx.x & 1) == 0) ? z : w; //zw000000000000000000000000000000
      a += __SHFL_xor < TYPE > ( a,  1, 32 );
      b += __SHFL_xor < TYPE > ( b,  1, 32 );
      a = ((threadIdx.x & 1) == 0) ? a : b; //ab000000000000000000000000000000
      c += __SHFL_xor < TYPE > ( c,  1, 32 );
      d += __SHFL_xor < TYPE > ( d,  1, 32 );
      c = ((threadIdx.x & 1) == 0) ? c : d; //cd000000000000000000000000000000
      x += __SHFL_xor < TYPE > ( z,  2, 32);//xyzw0000000000000000000000000000
      a += __SHFL_xor < TYPE > ( c,  2, 32);//abcd0000000000000000000000000000
      x += __SHFL_xor < TYPE > ( a,  4, 32);//xyzwabcd000000000000000000000000
}

      return x;
}
#endif

template < class TYPE, int BLOCK_SIZE >
__forceinline__ __device__ void
reduce_in_block ( TYPE x0, TYPE * w, TYPE * z )
{

      int local_id = threadIdx.x;

//
// 1. w and z must have shread memory attribute
// 2. BLOCK_SIZE is assumed to be hold
//       32 <= BLOCK_SIZE <= 32 * 16
//       and a multiple of 32
//    or BLOCK_SIZE = 16
//

  if ( BLOCK_SIZE == 16 ) {

      x0 = half_warp_Reduce < TYPE > ( x0 );

  } else {

      x0 = warp_Reduce < TYPE, 32 > ( x0 );

  if ( BLOCK_SIZE > 32 ) {

      int t = local_id / 32;
      int k = local_id % 32;

      __syncthreads();
      if ( k < 1 ) { w[t] = x0; }
      __syncthreads();

      if ( local_id < 32 ) {
         x0 = ( k < BLOCK_SIZE / 32 ) ? w[k] : (TYPE)0;
         x0 = warp_Reduce < TYPE, BLOCK_SIZE/32 > ( x0 );
      }
  }

  }
      __syncthreads();
      if ( local_id == 0 ) z[0] = x0;

}

template < class TYPE, int BLOCK_SIZE >
__forceinline__ __device__ void
reduce2_in_block ( TYPE x0, TYPE x1, TYPE * w, TYPE * z )
{

      int local_id = threadIdx.x;

//
// 1. w must have shread memory attribute
// 2. BLOCK_SIZE is assumed to be hold
//       32 <= BLOCK_SIZE <= 32 * 16
//       and a multiple of 32
//    or BLOCK_SIZE = 16
//

  if ( BLOCK_SIZE == 16 ) {

      x0 = half_warp_Reduce_2 < TYPE > ( x0, x1 );

  } else {

      x0 = warp_Reduce_2 < TYPE, 32 > ( x0, x1 );


  if ( BLOCK_SIZE > 32 ) {

      int t = local_id / 32;
      int k = local_id % 32;
      int L = BLOCK_SIZE / 32;

      __syncthreads();
      if ( k < 2 ) { w[t+L*(k*2)] = x0; w[t+L*(k*2+1)] = (TYPE)0; }
      __syncthreads();

      if ( local_id < 32 ) {
         k = ((k < L) ? k : L);
         x0 = w[k+0*2*L];
         x1 = w[k+1*2*L];
         x0 = warp_Reduce_2 < TYPE, BLOCK_SIZE/32 > ( x0, x1 );
      }

  }

  }
      __syncthreads();
      if ( local_id < 2 ) z[local_id] = x0;
}

template < class TYPE, int BLOCK_SIZE >
__forceinline__ __device__ void
reduce3_in_block ( TYPE x0, TYPE x1, TYPE x2, TYPE * w, TYPE * z )
{

      int local_id = threadIdx.x;

//
// 1. w must have shread memory attribute
// 2. BLOCK_SIZE is assumed to be hold
//       32 <= BLOCK_SIZE <= 32 * 16
//       and a multiple of 32
//    or BLOCK_SIZE = 16
//

  if ( BLOCK_SIZE == 16 ) {

      x0 = half_warp_Reduce_3 < TYPE > ( x0, x1, x2 );

  } else {

      x0 = warp_Reduce_3 < TYPE, 32 > ( x0, x1, x2 );

  if ( BLOCK_SIZE > 32 ) {

      int t = local_id / 32;
      int k = local_id % 32;
      int L = BLOCK_SIZE / 32;

      __syncthreads();
      if ( k < 3 ) { w[t+L*(k*2)] = x0; w[t+L*(k*2+1)] = (TYPE)0; }
      __syncthreads();

      if ( local_id < 32 ) {
         k = ((k < L) ? k : L);
         x0 = w[k+0*2*L];
         x1 = w[k+1*2*L];
         x2 = w[k+2*2*L];
         x0 = warp_Reduce_3 < TYPE, BLOCK_SIZE/32 > ( x0, x1, x2 );
      }

  }

  }
      __syncthreads();
      if ( local_id < 3 ) z[local_id] = x0;
}

template < class TYPE, int BLOCK_SIZE >
__forceinline__ __device__ void
reduce4_in_block ( TYPE x0, TYPE x1, TYPE x2, TYPE x3, TYPE * w, TYPE * z )
{

      int local_id = threadIdx.x;

//
// 1. w must have shread memory attribute
// 2. BLOCK_SIZE is assumed to be hold
//       32 <= BLOCK_SIZE <= 32 * 16
//       and a multiple of 32
//    or BLOCK_SIZE = 16
//

  if ( BLOCK_SIZE == 16 ) {

      x0 = half_warp_Reduce_4 < TYPE > ( x0, x1, x2, x3 );

  } else {

      x0 = warp_Reduce_4 < TYPE, 32 > ( x0, x1, x2, x3 );

  if ( BLOCK_SIZE > 32 ) {

      int t = local_id / 32;
      int k = local_id % 32;
      int L = BLOCK_SIZE / 32;

      __syncthreads();
      if ( k < 4 ) { w[t+L*(k*2)] = x0; w[t+L*(k*2+1)] = (TYPE)0; }
      __syncthreads();

      if ( local_id < 32 ) {
         k = ((k < L) ? k : L);
         x0 = w[k+0*2*L];
         x1 = w[k+1*2*L];
         x2 = w[k+2*2*L];
         x3 = w[k+3*2*L];
         x0 = warp_Reduce_4 < TYPE, BLOCK_SIZE/32 > ( x0, x1, x2, x3 );
      }

  }

  }
      __syncthreads();
      if ( local_id < 4 ) z[local_id] = x0;
}

#if 0
template < class TYPE, int BLOCK_SIZE >
__forceinline__ __device__ void
reduce8_in_block ( TYPE x0, TYPE x1, TYPE x2, TYPE x3, TYPE x4, TYPE x5, TYPE x6, TYPE x7, TYPE * w, TYPE * z )
{

      int local_id = threadIdx.x;

//
// 1. w must have shread memory attribute
// 2. BLOCK_SIZE is assumed to be hold
//       32 <= BLOCK_SIZE <= 32 * 16
//       and a multiple of 32
//    or BLOCK_SIZE = 16
//

  if ( BLOCK_SIZE == 16 ) {

      x0 = half_warp_Reduce_8 < TYPE > ( x0, x1, x2, x3, x4, x5, x6, x7 );

  } else {

      x0 = warp_Reduce_8 < TYPE, 32 > ( x0, x1, x2, x3, x4, x5, x6, x7 );

  if ( BLOCK_SIZE > 32 ) {

      int t = local_id / 32;
      int k = local_id % 32;
      int L = BLOCK_SIZE / 32;

      __syncthreads();
      if ( k < 8 ) { w[t+L*(k*2)] = x0; w[t+L*(k*2+1)] = (TYPE)0; }
      __syncthreads();

      if ( local_id < 32 ) {
         k = ((k < L) ? k : L);
         x0 = w[k+0*2*L];
         x1 = w[k+1*2*L];
         x2 = w[k+2*2*L];
         x3 = w[k+3*2*L];
         x4 = w[k+4*2*L];
         x5 = w[k+5*2*L];
         x6 = w[k+6*2*L];
         x7 = w[k+7*2*L];
         x0 = warp_Reduce_8 < TYPE, BLOCK_SIZE/32 > ( x0, x1, x2, x3, x4, x5, x6, x7 );
      }

  }

  }
      __syncthreads();
      if ( local_id < 8 ) z[local_id] = x0;
}
#endif

#endif


