#!/bin/bash

Init()
{

	/bin/sh ../get_dev_info.sh > DEV_INFO

	DEVICE=`awk '/DEVICE=/{ print $2 }' DEV_INFO`
	MP=`awk '/MP=/{ print $2 }' DEV_INFO`
	CG=`awk '/CG=/{ print $2 }' DEV_INFO`
	MAXDIM=`awk '/MAXDIM=/{ print $2 }' DEV_INFO`

        \rm DEV_INFO

	if [ -e ../CURRENT_GPU ]; then
		\rm ../CURRENT_GPU
	fi
	M=`expr $CG "/" 100`
	echo '#define CURRENT_GPU '$M > ../CURRENT_GPU

	if [ -f IN-exe-a ]; then
		\rm IN-exe-a
	fi
	cat << EOS_DO_A >> IN-exe-a
	1
	128
	256
	512
	1001
	2001
	4001
	7001
	10001
EOS_DO_A
	N=13000
	N_STEP=3000
	while [ $N -lt $MAXDIM ]
	do
		echo $N >> IN-exe-a
		D_STEP=`expr $N_STEP "*" 33`
		D_STEP=`expr $D_STEP "/" 100`
		N_STEP=`expr $N_STEP "+" $D_STEP`
		N=`expr $N "+" $N_STEP`
	done
		echo $MAXDIM >> IN-exe-a 
	echo 0 >> IN-exe-a

	if [ -f IN-exe-b ]; then
		\rm IN-exe-b
	fi
cat << EOS_DO_B >> IN-exe-b
        1
        50
        50
        100
        150
        200
        250
        300
        400
        512
        600
        700
        800
        900
        950
        1001
        1201
        1401
        1601
        1901
        2201
        2301
        2401
        2501
        2601
        2701
        2801
        2901
        3001
        3101
        3201
        3401
        3501
        3601
        3701
        3801
        3901
        4001
        4501
        5001
        5501
        6001
        6501
        7001
        7501
        8001
        8501
        9001
        9501
        10001
EOS_DO_B
	N=12000
	while [ $N -lt $MAXDIM ]
	do
		echo $N >> IN-exe-b
		N=`expr $N "+" 2000`
	done
		echo $MAXDIM >> IN-exe-b 
	echo 0 >> IN-exe-b

	UX_PATTERN=" 6 7 8 9 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 32 33 34 35 36 37 38 39 40 41 42 43 44 45 46 47 48 49 50 51 52 53 54 55 56 57 58 59 60 61 62 63 64 65 66 67 68 69 70 71 72 73 74 75 76 77 78 79 80 "
	if [ $CG -lt 300 ]; then
		MM_PATTERN=" 8 7 6 5 4 3 2 "
		MM_MAX=8
	else
		MM_PATTERN=" 16 15 14 13 12 11 10 9 8 7 6 5 4 3 2 "
		MM_MAX=16
	fi
}

Setup()
{
	if [ -f dsymv-upper-auto.h ]; then
		\rm dsymv-upper-auto.h
	fi
	awk ' \
		BEGIN{ \
			print "switch (BLK) {"; \
			for ( i=1; i <= 10*'$MM_MAX'; i++ ) { \
				print "#if __KERNEL"i; \
				print "case "i":"; \
				print "\tAPI(SYMVu_ATOMIC_host) < GPU_ARCH, scalar_t, BLOCK_SIZE, VX, UX, "int((i-1)/10)+1", "(i-1)%10" >"; \
				print "\t( n, a, lda, x, y, alpha, beta ); break;"; \
				print "#endif" \
			} \
			print "case 0:"; \
			print "default:"; \
			print "\tAPI(DSYMVu_lu) "; \
			print "\t( n, a, lda, x, y, alpha, beta ); break;"; \
			print "}"; \
			exit; \
		} { exit; } END { exit; }' > dsymv-upper-auto.h
	cp dsymv-upper-auto.h ../
}

Make0()
{
	if [ -f param-dsymv.h ]; then
		\rm param-dsymv.h
	fi
	touch param-dsymv.h

	echo "#define BLOCK_SIZE "$BLOCK_SIZE >> param-dsymv.h
	echo "#define VX "$VX >> param-dsymv.h
	echo "#define UX "$UX >> param-dsymv.h

	echo "//"$MMM_PATTERN >> param-dsymv.h

	for MULTI in \
		$MMM_PATTERN
	do
	for MX in \
		0 1 2 3 4 5 6 7 8 9
	do
		ID=`expr $MULTI "-" 1`
		ID=`expr $ID "*" 10`
		ID=`expr $ID "+" $MX`
		ID=`expr $ID "+" 1`
		echo "#define	__KERNEL"$ID" 1" >> param-dsymv.h
	done
	done
	cp param-dsymv.h ../

}

Make()
{
	cd ../../src
	\rm dsymv_upper-*_template.cu_o
	make -f Makefile.at
	cd ../bench
	\rm test.o test2.o
	make test-d
	cd ../tuning/dsymv-current
}

Do()
{
	echo "#define BLOCK_SIZE "$BLOCK_SIZE
	echo "#define VX "$VX
	echo "#define UX "$UX
	echo "#define MULTIPLICITY "$MULTI
	echo "#define _M_ "$MX

	MX1=`expr $MX + 1`
	MX2=`expr $MULTI - 1`
	MX2=`expr $MX2 "*" 10`
	MX3=`expr $MX1 + $MX2`

	echo 1    >  IN-nv
	echo 1001 >> IN-nv
	echo 6001 >> IN-nv
	echo 9001 >> IN-nv
	echo -1   >> IN-nv

	ulimit -t 120

	../../bench/test-dsymv-u IN-nv $MX3 | \
		awk ' \
			/cannot free/{ \
				system("touch ./.NG"); \
			} \
			/ERR/{ \
				system("touch ./.NG"); \
			} \
			{ \
				print;fflush(); \
			} \
		' > /dev/null

	if [ ! -e ./.NG ]; then

		../../bench/test2-dsymv-u IN-exe-a $MX3

	fi
}



Init

Setup

echo "Sampling start"

for VX in \
	1 2 3 4 5 6 7 8
do
for BLOCK_SIZE in \
	32 64 96 128 160 192 224 256 288
do
if [ `expr $BLOCK_SIZE "*" $VX` -ge 96 ]; then
if [ `expr $BLOCK_SIZE "*" $VX` -le 288 ]; then

  touch log-regs-$BLOCK_SIZE
  \rm log-regs-$BLOCK_SIZE

  for UX in \
	$UX_PATTERN
  do
  if [ `expr $UX "%" $VX` -eq 0 ]; then
  if [ `expr $UX "/" $VX` -ge 4 ]; then
  if [ `expr $UX "/" $VX` -le 32 ]; then


	MMM_PATTERN="1"
	Make0 >& /dev/null

	M=`expr $UX "/" 2`
	M=`expr $M "*" 2`
	M=`expr $M "+" 1`
	M=`expr $M "*" $UX`
	M2=`expr $BLOCK_SIZE "*" $VX`
	if [ $M -lt $M2 ]; then
		M=$M2
	fi
	M1=`expr $M "%" 2`
	M=`expr $M "+" $M1`
	M=`expr $M "+" $UX`
	SHMEM_MIN=`expr $M "*" 8`

	/bin/sh ../get_reguse.sh $CG \
		../../src/dsymv_upper-Tesla_template \
		../../src/dsymv_upper-Fermi_template \
		../../src/dsymv_upper-Kepler_template \
		../../src/dsymv_upper-Kepler_template \
		../../src/dsymv_upper-Maxwell_template \
		main_do | \
        sort -k 5 > log-reg-all

	MMM=0
	CCC=0
	SSS=0
      for MX in \
	0 1 2 3 4 5 6 7 8 9
      do

	touch log-regs
	\rm log-regs

	echo 1 > IN-1
	echo 6001 >> IN-1
	echo -1 >> IN-1
	MX3=`expr $MX + 1`

	cat log-reg-all | awk '{if($5=='$MX')print $1}' > CX
	CX=`cat CX`

	cat log-reg-all | awk '{if($5=='$MX')print $7}' > SX
	SX=`cat SX`

	M=`expr $BLOCK_SIZE "*" $VX`
#	MMM_PATTERN=`/bin/sh ../get_mult.sh $M $CX $SHMEM_MIN $CG | \
	MMM_PATTERN=`/bin/sh ../get_mult48.sh $M $CX $SHMEM_MIN $CG | \
		awk '{ gsub(/\([0-9]*\)/," "); print; }'`
	M=`echo $MMM_PATTERN | awk '{ print $(NF); }'`
	if [ $M -gt $MMM ]; then
		if [ ! $MMM -eq 0 ]; then
			CCC=1
		fi
		MMM=$M
        else
		if [ ! $M -eq $MMM ]; then
			CCC=1
		fi
	fi

	if [ $SSS -lt $SX ]; then
		SSS=$SX
	fi

	echo $BLOCK_SIZE' '$VX' '$UX' '$MX' Register= '$CX ' / '$SX' :: '$SHMEM_MIN' : '$MMM_PATTERN | \
		tee -a log-regs-$BLOCK_SIZE

      done # for MX

	M=`expr $BLOCK_SIZE "/" 32`
	M=`expr $M "*" $VX`
	M=`expr $M "*" $MMM`
	M=`expr $M "*" 100`
	Occupancy=`expr $M "/" 64`

      echo '## '$BLOCK_SIZE' '$VX' '$UX' '$MMM' / '$Occupancy'% ! '$CCC' > '$SSS | \
		tee -a log-regs-$BLOCK_SIZE

        if [ $MMM -eq 1 ]; then
		break
	fi

  fi
  fi
  fi
  done # for UX

  DONE=0
  for MULTI in \
	$MM_PATTERN
  do

	if [ $DONE -gt 6 ]; then
		break
	fi

	MM16_PATTERN=" 16 12 10 9 8 7 6 5 4 3 2"
	MM48_PATTERN=" 16 14 13 12 11 10 9 8 7 6 5 4 3 2"

	MMxx_PATTERN=$MM48_PATTERN

	M_C=0
	for M in \
		$MMxx_PATTERN
	do
		if [ $M -eq $MULTI ]; then
			M_C=`expr $M_C "+" 1`
		fi
	done

	MMM_PATTERN=$MULTI

	if [ $M_C -eq 0 ]; then
	cat log-regs-$BLOCK_SIZE | \
		awk '/##/{ if ( $5 == '$MULTI' && $11 ) print $4 }' | \
		sort -r -g | \
		head -8 | awk '{printf("%d ",$0) }' > UUU
	cat log-regs-$BLOCK_SIZE | \
		awk '/##/{ if ( $5 == '$MULTI' && $11 ) print $9 }' | \
		sort -r -g | \
		head -1 | awk '{printf("%d ",$0) }' > CCC
	cat log-regs-$BLOCK_SIZE | \
		awk 'BEGIN{print 1}/##/{ if ( $5 == '$MULTI' && $11 ) print $11 }' | \
		sort -r -g | \
		head -1 | awk '{printf("%d ",$0) }' > SSS
	else
	cat log-regs-$BLOCK_SIZE | \
		awk '/##/{ if ( $5 >= '$MULTI' && $11 ) print $4 }' | \
		sort -r -g | \
		head -8 | awk '{printf("%d ",$0) }' > UUU
	cat log-regs-$BLOCK_SIZE | \
		awk '/##/{ if ( $5 >= '$MULTI' && $11 ) print $9 }' | \
		sort -r -g | \
		head -1 | awk '{printf("%d ",$0) }' > CCC
	cat log-regs-$BLOCK_SIZE | \
		awk 'BEGIN{print 1}/##/{ if ( $5 >= '$MULTI' && $11 ) print $11 }' | \
		sort -r -g | \
		head -1 | awk '{printf("%d ",$0) }' > SSS
	fi
	UU_PATTERN=`cat UUU`
	CCC=`cat CCC`
	SSS=`cat SSS`

	if [ $SSS -gt 0 ]; then
	if [ $SSS -gt 1 ]; then
		CCC=3
	else
		CCC=1
	fi
	fi

	if [ x${UU_PATTERN} = x ]; then
		echo 'UX['$MULTI']:=<none>'
	else
		M=`expr $CCC "+" 1`
		echo 'UX['$MULTI']:=<'$UU_PATTERN'> <= '$M
	fi
	
    for UX in \
	$UU_PATTERN
    do
    if [ $UX -le $BLOCK_SIZE ]; then
    if [ `expr $UX "/" $VX` -ge 4 ]; then
    if [ `expr $UX "/" $VX` -le 32 ]; then

	Make0 >& /dev/null
	Make  >& /dev/null

		if [ -e NG ]; then
			\rm NG
		fi

      for MX in \
	0 1 2 3 4 5 6 7 8 9
      do

		if [ -e .NG ]; then
			\rm .NG
		fi

		Do | awk ' \
			/ERROR/ { \
				print; \
				fflush(); \
				system("touch NG"); \
				exit; \
				next; \
			} \
			{ \
				print; \
				fflush(); \
			} \
			END { \
				fflush(); \
			} \
			'

		if [ -e .NG ]; then
			echo 'ERROR in kernel implementation!!'
			\rm .NG
			touch NG
		fi

		if [ -e NG ]; then
			break
		fi

      done # for MX

		if [ -e NG ]; then
			\rm NG
		else
			if [ $CCC -gt 0 ]; then
				CCC=`expr $CCC "-" 1`
			else
				DONE=`expr $DONE "+" 1`
				break
			fi
		fi

    fi
    fi
    fi
    done # for UX

  done # for MULTI

fi
fi
done # for BLOCK_SIZE
done # VX

echo "Complete"

